"""Applications interface

   Access to the applications endpoint.

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      application = arch.applications.create(...)

"""

from logging import getLogger
from typing import TYPE_CHECKING, Any

# pylint:disable=cyclic-import      # but pylint doesn't understand this feature
if TYPE_CHECKING:
    from .archivist import Archivist

from .constants import (
    APPLICATIONS_LABEL,
    APPLICATIONS_REGENERATE,
    APPLICATIONS_SUBPATH,
)
from .dictmerge import _deepmerge

LOGGER = getLogger(__name__)


class Application(dict):
    """Application object"""


class _ApplicationsClient:
    """ApplicationsClient

    Access to applications entities using CRUD interface. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._subpath = f"{archivist_instance.root}/{APPLICATIONS_SUBPATH}"
        self._label = f"{self._subpath}/{APPLICATIONS_LABEL}"

    def __str__(self) -> str:
        return f"ApplicationsClient({self._archivist.url})"

    def create(self, display_name: str, custom_claims: "dict[str, str]") -> Application:
        """Create application

        Creates application with defined attributes.

        Args:
            display_name (str): display name of application.
            custom_claims (dict): custom claims

        Returns:
            :class:`Application` instance

        """
        LOGGER.debug("Create Application %s", display_name)
        return self.create_from_data(
            self.__params(
                display_name=display_name,
                custom_claims=custom_claims,
            ),
        )

    def create_from_data(self, data: "dict[str, Any]") -> Application:
        """Create application

        Creates application with request body from data stream.
        Suitable for reading data from a file using json.load or yaml.load

        Args:
            data (dict): request body of application.

        Returns:
            :class:`Application` instance

        """
        return Application(**self._archivist.post(self._label, data))

    def read(self, identity: str) -> Application:
        """Read Application

        Reads application.

        Args:
            identity (str): applications identity e.g. applications/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            :class:`Application` instance

        """
        return Application(**self._archivist.get(f"{self._subpath}/{identity}"))

    def update(
        self,
        identity: str,
        *,
        display_name: "str|None" = None,
        custom_claims: "dict[str,str]|None" = None,
    ) -> Application:
        """Update Application

        Update application.

        Args:
            identity (str): applications identity e.g. applications/xxxxxxxxxxxxxxxxxxxxxxx
            display_name (str): display name of application.
            custom_claims (dict): custom claims

        Returns:
            :class:`Application` instance

        """
        return Application(
            **self._archivist.patch(
                f"{self._subpath}/{identity}",
                self.__params(
                    display_name=display_name,
                    custom_claims=custom_claims,
                ),
            )
        )

    def delete(self, identity: str) -> "dict[str, Any]":
        """Delete Application

        Deletes application.

        Args:
            identity (str): applications identity e.g. applications/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            :class:`Application` instance - empty?

        """
        return self._archivist.delete(f"{self._subpath}/{identity}")

    def __params(
        self,
        *,
        display_name: "str|None" = None,
        custom_claims: "dict[str,str] | None" = None,
    ) -> "dict[str, Any]":
        params = {}

        if display_name is not None:
            params["display_name"] = display_name

        if custom_claims is not None:
            params["custom_claims"] = custom_claims

        return _deepmerge(self._archivist.fixtures.get(APPLICATIONS_LABEL), params)

    def list(
        self,
        *,
        page_size: "int|None" = None,
        display_name: "str|None" = None,
    ):
        """List applications.

        List applications that match criteria.

        Args:
            display_name (str): display name (optional)
            page_size (int): optional page size. (Rarely used).

        Returns:
            iterable that returns :class:`Application` instances

        """
        return (
            Application(**a)
            for a in self._archivist.list(
                self._label,
                APPLICATIONS_LABEL,
                page_size=page_size,
                params=self.__params(display_name=display_name),
            )
        )

    def regenerate(self, identity: str) -> Application:
        """Regenerate secret

        Regenerate secret for app registration

        Args:
            identity (str): identity of application

        Returns:
            :class:`Application` instance

        """
        LOGGER.debug("Regenerate %s", identity)
        return Application(
            **self._archivist.post(
                f"{self._subpath}/{identity}:{APPLICATIONS_REGENERATE}",
                None,
            )
        )



"""Archivist Compliance Policy Type

   Enumerated type that allows user to select the compliance policy type when
   creating a compliance policy.

"""

from enum import Enum


class CompliancePolicyType(Enum):
    """
    Enumerate types of compliance policy
    """

    COMPLIANCE_TYPE_UNDEFINED = 0
    #: Time since specific event for specified period
    COMPLIANCE_SINCE = 1
    #: Unresolved event currently on asset (e.g. vulnerability)
    COMPLIANCE_CURRENT_OUTSTANDING = 2
    #: No unresolved events for longer than specified period
    COMPLIANCE_PERIOD_OUTSTANDING = 3
    #: dynamic tolerance with dynamic window etc..
    COMPLIANCE_DYNAMIC_TOLERANCE = 4
    #: Compliance on comparison of asset attribute value to predefined comparator
    COMPLIANCE_RICHNESS = 5



"""Attachments interface

   Direct access to the attachments endpoint

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      with open("something.jpg") as fd:
          attachment = arch.attachments.upload(fd)

"""

# pylint:disable=too-few-public-methods


from copy import deepcopy
from logging import getLogger
from typing import TYPE_CHECKING, Any, BinaryIO
from urllib.parse import urlparse

if TYPE_CHECKING:
    from requests.models import Response

    # pylint:disable=cyclic-import      # but pylint doesn't understand this feature
    from .archivist import Archivist

from .constants import (
    ASSETATTACHMENTS_LABEL,
    ASSETATTACHMENTS_SUBPATH,
    ATTACHMENTS_LABEL,
    SEP,
)
from .dictmerge import _deepmerge

LOGGER = getLogger(__name__)


class _AssetAttachmentsClient:
    """AssetAttachmentsClient

    Access to attachments entities using CRUD interface. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._public = archivist_instance.public
        self._subpath = f"{archivist_instance.root}/{ASSETATTACHMENTS_SUBPATH}"
        self._label = f"{self._subpath}/{ASSETATTACHMENTS_LABEL}"

    def __str__(self) -> str:
        if self._public:
            return "AssetAttachmentsClient()"

        return f"AssetAttachmentsClient({self._archivist.url})"

    def _identity(self, identity: str, attachment_id: str) -> str:
        """Return fully qualified identity
        If public then expect a full url as argument

        identity looks like:

        [https://app.datatrails.ai/archivist/public]assets/xxxxxxx

        OR

        [https://app.datatrails.ai/archivist/public]assets/xxxxxxx/events/yyyyyy

        where the public URL is prefixed with the schema.

        """
        uuid = attachment_id.split(SEP)[1]
        if self._public:
            # the public URL for the asset or event has to be changed
            url = urlparse(identity)
            root = "/".join(url.path.split(SEP)[:2])
            asset_id = "/".join(url.path.split(SEP)[2:])
            new_url = url._replace(
                path=f"{root}/{ASSETATTACHMENTS_SUBPATH}/{ASSETATTACHMENTS_LABEL}/{asset_id}/{uuid}"
            )
            return new_url.geturl()

        return f"{self._label}/{identity}/{uuid}"

    def __params(self, params: "dict[str, Any]|None") -> "dict[str, Any]":
        params = deepcopy(params) if params else {}
        # pylint: disable=protected-access
        return _deepmerge(self._archivist.fixtures.get(ATTACHMENTS_LABEL), params)

    def download(
        self,
        identity: str,
        attachment_id: str,
        fd: BinaryIO,
        *,
        params: "dict[str, Any]|None" = None,
    ) -> "Response":
        """Read attachment

        Reads attachment into data sink (usually a file opened for write).
        Note that returns the response as the body will be consumed by the
        fd iterator

        Args:
            identity (str): identity
            attachment_id (str): blobs/aaaaaaaaaaaaa
            fd (file): opened file descriptor or other file-type sink.
            params (dict): e.g. {"allow_insecure": "true"} OR {"strict": "true" }

        Returns:
            JSON as dict

        identity has one of the following 4 forms:

            [https://app.datatrails.ai/archivist/][v2/] - if public
            assets/xxxxxxxxxxxxxxxxxxxxxxxxxx

            [https://app.datatrails.ai/archivist/][v2/] - if public
            assets/xxxxxxxxxxxxxxxxxxxxxxxxxx/events/yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy

        """

        return self._archivist.get_file(
            self._identity(identity, attachment_id), fd, params=self.__params(params)
        )

    def info(
        self,
        identity: str,
        attachment_id: str,
    ) -> "dict[str, Any]":
        """Read asset attachment info

        Reads asset attachment info

        Args:
            identity (str): identity
            attachment_id (str): blobs/aaaaaaaaaaaaa

        Returns:
            REST response

        identity has one of the following 4 forms:

            [https://app.datatrails.ai/archivist/][v2/] - if public
            assets/xxxxxxxxxxxxxxxxxxxxxxxxxx

            [https://app.datatrails.ai/archivist/][v2/] - if public
            assets/xxxxxxxxxxxxxxxxxxxxxxxxxx/events/yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy
        """

        return self._archivist.get(f"{self._identity(identity, attachment_id)}/info")



"""Archivist dict deep merge
"""


from copy import deepcopy
from typing import Any

from flatten_dict import flatten, unflatten


def _deepmerge(
    dct1: "dict[str, Any]|None", dct2: "dict[str, Any]|None"
) -> "dict[str, Any]":
    """Deep merge 2 dictionaries

    The settings from dct2 overwrite or add to dct1
    """
    if dct1 is None:
        if dct2 is None:
            return {}
        return deepcopy(dct2)

    if dct2 is None:
        return deepcopy(dct1)

    return unflatten({**flatten(dct1), **flatten(dct2)})


def _dotdict(dct: "dict[str, Any]|None") -> "dict[str, str] | None":
    """Emit nested dictionary as dot delimited dict with one level"""
    if dct is None:
        return None
    return flatten(dct, reducer="dot")



"""Creates the root logger as a simple console streamer. Follows the Django-style
   logging configuration where the logging system forms a hierarchy of loggers
   all potentially independently configurable.

   This model allows controlling loggers that are part of the dependency list. For
   example setting a debug logging level will show debug output from the underlying
   urllib3 package.

   URL: https://docs.djangoproject.com/en/3.2/topics/logging/
"""

# pylint:  disable=missing-docstring

from logging import config, getLogger

# root logger for all code
LOGGER = getLogger()


def set_logger(level):
    """
    Set logging level

    Args:
        level (str): Either INFO or DEBUG

    """
    config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "propagate": True,
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                },
            },
            "root": {
                "handlers": ["console"],
                "level": level,
            },
        }
    )



"""Appidp interface

   Access to the Appidp endpoint.

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      appidp = arch.appidp.token(...)

"""

from logging import getLogger
from typing import TYPE_CHECKING

# pylint:disable=cyclic-import      # but pylint doesn't understand this feature
# pylint:disable=too-few-public-methods

# pylint:disable=cyclic-import      # but pylint doesn't understand this feature
if TYPE_CHECKING:
    from .archivist import Archivist

from .constants import (
    APPIDP_LABEL,
    APPIDP_SUBPATH,
    APPIDP_TOKEN,
)

LOGGER = getLogger(__name__)


class AppIDP(dict):
    """Appidp object"""


class _AppIDPClient:
    """AppIDP Client

    Access to appidp entities. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._subpath = f"{archivist_instance.root}/{APPIDP_SUBPATH}"
        self._label = f"{self._subpath}/{APPIDP_LABEL}"

    def __str__(self) -> str:
        return f"AppIDPClient({self._archivist.url})"

    def token(self, client_id: str, client_secret: str) -> AppIDP:
        """Create access token from client id and secret

        Args:
            client_id (str): client id
            client_secret (str): client_secret

        Returns:
            :class:`AppIDP` instance

        """
        return AppIDP(
            **self._archivist.post(
                f"{self._label}/{APPIDP_TOKEN}",
                {
                    "grant_type": "client_credentials",
                    "client_id": client_id,
                    "client_secret": client_secret,
                },
                data=True,
            )
        )



# pylint:  disable=missing-docstring

from logging import getLogger
from sys import exit as sys_exit
from sys import stdout as sys_stdout

from ...parser import common_parser, endpoint
from .run import run

LOGGER = getLogger(__name__)


def main():
    parser = common_parser("Executes the archivist runner from a yaml file")

    parser.add_argument(
        "yamlfile", help="the yaml file describing the operations to conduct"
    )
    args = parser.parse_args()

    arch = endpoint(args)

    run(arch, args)

    parser.print_help(sys_stdout)
    sys_exit(1)



"""Archivist Proof Mechanism

   Enumerated type that allows user to select the proof_mechanism option when
   creating an asset.

"""

# pylint: disable=unused-private-member

from enum import Enum


class ProofMechanism(Enum):
    """Enumerate proof mechanism options"""

    # previously used but now removed
    __RESERVED = 1
    #: Assets and events are proven using a hash of the originator's evidence
    SIMPLE_HASH = 2



"""Compliance Policies interface

   Access to the compliance_policies endpoint.

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )

      # A 'Since' policy
      asset = arch.compliance_policies.create(
          ComplianceTypeSince(...)
      )

"""


from copy import deepcopy
from logging import getLogger
from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    # pylint:disable=cyclic-import      # but pylint doesn't understand this feature
    from .archivist import Archivist
    from .compliance_policy_requests import (
        CompliancePolicyCurrentOutstanding,
        CompliancePolicyDynamicTolerance,
        CompliancePolicyPeriodOutstanding,
        CompliancePolicyRichness,
        CompliancePolicySince,
    )

from .constants import (
    COMPLIANCE_POLICIES_LABEL,
    COMPLIANCE_POLICIES_SUBPATH,
)
from .dictmerge import _deepmerge

LOGGER = getLogger(__name__)


class CompliancePolicy(dict):
    """CompliancePolicy

    CompliancePolicy object has dictionary of all the compliance policy attributes.

    """

    @property
    def name(self):
        """str: name of the compliance policy"""
        return self.get("display_name")


class _CompliancePoliciesClient:
    """CompliancePoliciesClient

    Access to compliance policy entities using CRUD interface. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._subpath = f"{archivist_instance.root}/{COMPLIANCE_POLICIES_SUBPATH}"
        self._label = f"{self._subpath}/{COMPLIANCE_POLICIES_LABEL}"

    def __str__(self) -> str:
        return f"CompliancePoliciesClient({self._archivist.url})"

    def create(
        self,
        policy: Union[
            "CompliancePolicySince",
            "CompliancePolicyCurrentOutstanding",
            "CompliancePolicyPeriodOutstanding",
            "CompliancePolicyDynamicTolerance",
            "CompliancePolicyRichness",
        ],
    ) -> CompliancePolicy:
        """Create A compliance policy

        Args:
            policy (CompliancePolicy): the policy object.
            One of:
                CompliancePolicySince
                CompliancePolicyCurrentOutstanding
                CompliancePolicyPeriodOutstanding
                CompliancePolicyDynamicTolerance
                CompliancePolicyRichness

        Returns:
            :class:`CompliancePolicy` instance

        """
        return self.create_from_data(policy.dict())

    def create_from_data(self, data: "dict[str, Any]") -> "CompliancePolicy":
        """Create compliance_policy

        Creates compliance_policy with request body from data stream.
        Suitable for reading data from a file using json.load or yaml.load

        Args:
            data (dict): request body of compliance_policy.

        Returns:
            :class:`CompliancePolicy` instance

        """
        return CompliancePolicy(**self._archivist.post(self._label, data))

    def read(self, identity: str) -> CompliancePolicy:
        """Read compliance policy

        Reads compliance policy.

        Args:
            identity (str): compliance policy identity
                            e.g. compliance_policies/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            :class:`CompliancePolicy` instance

        """
        return CompliancePolicy(**self._archivist.get(f"{self._subpath}/{identity}"))

    def delete(self, identity: str) -> "dict[str, Any]":
        """Delete Compliance Policy

        Deletes compliance policy.

        Args:
            identity (str): compliance policy identity
            e.g. compliance_policies/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            :class:`CompliancePolicy` instance - empty?

        """
        return self._archivist.delete(f"{self._subpath}/{identity}")

    def __params(self, props: "dict[str, Any]|None") -> "dict[str, Any]":
        params = deepcopy(props) if props else {}
        # pylint: disable=protected-access
        return _deepmerge(
            self._archivist.fixtures.get(COMPLIANCE_POLICIES_LABEL), params
        )

    def count(self, *, props: "dict[str, Any]|None" = None) -> int:
        """Count compliance policies.

        Counts number of compliance policies that match criteria.

        Args:
            props (dict): e.g. {"compliance_type": "COMPLIANCE_RICHNESS" }

        Returns:
            integer count of compliance policies.

        """
        return self._archivist.count(
            self._label,
            params=self.__params(props),
        )

    def list(
        self, *, page_size: "int|None" = None, props: "dict[str, Any]|None" = None
    ):
        """List compliance policies.

        Lists compliance policies that match criteria.

        Args:
            props (dict): optional e.g. {"compliance_type": "COMPLIANCE_DYNAMIC_TOLERANCE" }
            page_size (int): optional page size. (Rarely used).

        Returns:
            iterable that returns :class:`CompliancePolicy` instances

        """
        return (
            CompliancePolicy(**a)
            for a in self._archivist.list(
                self._label,
                COMPLIANCE_POLICIES_LABEL,
                page_size=page_size,
                params=self.__params(props),
            )
        )

    def read_by_signature(self, *, props: "dict[str, Any]|None" = None):
        """Read compliance policy by signature.

        Reads compliance policy that meets criteria. Only one compliance policy is expected.

        Args:
            props (dict): e.g. {"display_name": "foo" }

        Returns:
            :class:`CompliancePolicy` instance

        """
        return CompliancePolicy(
            **self._archivist.get_by_signature(
                self._label,
                COMPLIANCE_POLICIES_LABEL,
                params=self.__params(props),
            )
        )



"""Subjects interface

   Access to the subjects endpoint.

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      subject = arch.subjects.create(...)

"""


from base64 import b64decode
from json import loads as json_loads
from logging import getLogger
from typing import TYPE_CHECKING, Any

# pylint:disable=cyclic-import      # but pylint doesn't understand this feature
from . import subjects_confirmer
from .constants import (
    SUBJECTS_LABEL,
    SUBJECTS_SELF_ID,
    SUBJECTS_SUBPATH,
)
from .dictmerge import _deepmerge

if TYPE_CHECKING:
    from .archivist import Archivist

LOGGER = getLogger(__name__)


class Subject(dict):
    """Subject object"""


class _SubjectsClient:
    """SubjectsClient

    Access to subjects entities using CRUD interface. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    maxDiff = None

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._subpath = f"{archivist_instance.root}/{SUBJECTS_SUBPATH}"
        self._label = f"{self._subpath}/{SUBJECTS_LABEL}"

    def __str__(self) -> str:
        return f"SubjectsClient({self._archivist.url})"

    def create(
        self,
        display_name: str,
        wallet_pub_key: "list[str]",
        tessera_pub_key: "list[str]",
    ) -> Subject:
        """Create subject

        Creates subject with defined attributes.

        Args:
            display_name (str): display name of subject.
            wallet_pub_key (list): wallet public keys
            tessera_pub_key (list): tessera public keys

        Returns:
            :class:`Subject` instance

        """
        LOGGER.debug("Create Subject %s", display_name)
        return self.create_from_data(
            self.__params(
                display_name=display_name,
                wallet_pub_key=wallet_pub_key,
                tessera_pub_key=tessera_pub_key,
            ),
        )

    def share(
        self, name: str, other_name: str, other_archivist: "Archivist"
    ) -> "tuple[Subject, Subject]":
        """Import the self subjects from the foreign archivist connection
           from another organization - mutually share.

        Args:
            name (str): display_name of the foreign self subject in this archivist
            other_name (str): display_name of the self subject in other archivist
            other_archivist (Archivist): Archivist object

        Returns:
            2-tuple of :class:`Subject` instance

        """
        subject1 = self.import_subject(
            name, other_archivist.subjects.read(SUBJECTS_SELF_ID)
        )
        subject2 = other_archivist.subjects.import_subject(
            other_name, self.read(SUBJECTS_SELF_ID)
        )
        subject1 = self.wait_for_confirmation(subject1["identity"])
        subject2 = other_archivist.subjects.wait_for_confirmation(subject2["identity"])

        return subject1, subject2

    def import_subject(self, display_name: str, subject: Subject) -> Subject:
        """Create subject from another subject usually
           from another organization.

        Args:
            display_name (str): display_name of the subject
            subject (Subject): Subject object

        Returns:
            :class:`Subject` instance

        """
        return self.create(
            display_name,
            subject["wallet_pub_key"],
            subject["tessera_pub_key"],
        )

    def create_from_data(self, data: "dict[str, Any]") -> Subject:
        """Create subject

        Creates subject with request body from data stream.
        Suitable for reading data from a file using json.load or yaml.load

        Args:
            data (dict): request body of subject.

        Returns:
            :class:`Subject` instance

        """
        LOGGER.debug("Create Subject from data %s", data)
        return Subject(**self._archivist.post(self._label, data))

    def create_from_b64(self, data: "dict[str, Any]") -> Subject:
        """Create subject

        Creates subject with request body from b64 encoded string

        Args:
            data (dict): Dictionary with 2 fields:

        A YAML representation of the data argument would be:

            .. code-block:: yaml

                display_name: An imported subject
                subject_string: ey66...

        Returns:
            :class:`Subject` instance

        """
        decoded = b64decode(data["subject_string"])
        LOGGER.debug("decoded %s", decoded)
        outdata = {
            k: v
            for k, v in json_loads(decoded).items()
            if k in ("wallet_pub_key", "tessera_pub_key")
        }
        outdata["display_name"] = data["display_name"]
        LOGGER.debug("data %s", outdata)

        return Subject(**self._archivist.post(self._label, outdata))

    def wait_for_confirmation(self, identity: str) -> Subject:
        """Wait for subject to be confirmed.

        Waits for subject to be confirmed.

        Args:
            identity (str): identity of asset

        Returns:
            True if subject is confirmed.

        """
        subjects_confirmer.MAX_TIME = self._archivist.max_time
        # pylint: disable=protected-access
        return subjects_confirmer._wait_for_confirmation(self, identity)

    def read(self, identity: str) -> Subject:
        """Read Subject

        Reads subject.

        Args:
            identity (str): subjects identity e.g. subjects/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            :class:`Subject` instance

        """
        return Subject(**self._archivist.get(f"{self._subpath}/{identity}"))

    def update(
        self,
        identity: str,
        *,
        display_name: "str|None" = None,
        wallet_pub_key: "list[str]|None" = None,
        tessera_pub_key: "list[str]|None" = None,
    ) -> Subject:
        """Update Subject

        Update subject.

        Args:
            identity (str): subjects identity e.g. subjects/xxxxxxxxxxxxxxxxxxxxxxx
            display_name (str): display name of subject.
            wallet_pub_key (list): wallet public keys
            tessera_pub_key (list): tessera public keys

        Returns:
            :class:`Subject` instance

        """
        return Subject(
            **self._archivist.patch(
                f"{self._subpath}/{identity}",
                self.__params(
                    display_name=display_name,
                    wallet_pub_key=wallet_pub_key,
                    tessera_pub_key=tessera_pub_key,
                ),
            )
        )

    def delete(self, identity: str) -> "dict[str, Any]":
        """Delete Subject

        Deletes subject.

        Args:
            identity (str): subjects identity e.g. subjects/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            :class:`Subject` instance - empty?

        """
        return self._archivist.delete(f"{self._subpath}/{identity}")

    def __params(
        self,
        *,
        display_name: "str|None" = None,
        wallet_pub_key: "list[str]|None" = None,
        tessera_pub_key: "list[str]|None" = None,
    ) -> "dict[str, Any]":
        params = {}

        if display_name is not None:
            params["display_name"] = display_name

        if wallet_pub_key is not None:
            params["wallet_pub_key"] = wallet_pub_key

        if tessera_pub_key is not None:
            params["tessera_pub_key"] = tessera_pub_key

        return _deepmerge(self._archivist.fixtures.get(SUBJECTS_LABEL), params)

    def count(self, *, display_name: "str|None" = None) -> int:
        """Count subjects.

        Counts number of subjects that match criteria.

        Args:
            display_name (str): display name (optional)

        Returns:
            integer count of subjects.

        """
        return self._archivist.count(
            self._label,
            params=self.__params(display_name=display_name),
        )

    def list(
        self,
        *,
        page_size: "int|None" = None,
        display_name: "str|None" = None,
    ):
        """List subjects.

        List subjects that match criteria.

        Args:
            display_name (str): display name (optional)
            page_size (int): optional page size. (Rarely used).

        Returns:
            iterable that returns :class:`Subject` instances

        """

        LOGGER.debug("List '%s'", display_name)
        return (
            Subject(**a)
            for a in self._archivist.list(
                self._label,
                SUBJECTS_LABEL,
                page_size=page_size,
                params=self.__params(display_name=display_name),
            )
        )



"""Archivist exceptions

   All exceptions are derived from a base ArchivistError class.

"""

import json
from logging import getLogger

from requests import Response

from .constants import HEADERS_RETRY_AFTER
from .headers import _headers_get

LOGGER = getLogger(__name__)


class ArchivistError(Exception):
    """Base exception for archivist package"""


class ArchivistBadFieldError(ArchivistError):
    """Incorrect field name in list() method"""


class ArchivistUnconfirmedError(ArchivistError):
    """asset or event failed to confirm after fixed timeout"""


class ArchivistInvalidOperationError(ArchivistError):
    """Runner Operation is invalid"""


class ArchivistBadRequestError(ArchivistError):
    """Ill-formed request or validation error (400)"""


class ArchivistDuplicateError(ArchivistError):
    """Read by signature returns more than one asset"""


class ArchivistUnauthenticatedError(ArchivistError):
    """user is unknown (401)"""


class ArchivistPaymentRequiredError(ArchivistError):
    """A quota has been reached (402)"""


class ArchivistForbiddenError(ArchivistError):
    """User does not have permission (403)"""


class ArchivistNotFoundError(ArchivistError):
    """Entity does not exist (404)"""


class ArchivistTooManyRequestsError(ArchivistError):
    """Too many requests in too short a time (429)"""

    def __init__(self, retry: "str|None", *args):
        self.retry = float(retry) if retry is not None else 0
        super().__init__(*args)


class Archivist4xxError(ArchivistError):
    """Any other 4xx error"""


class ArchivistNotImplementedError(ArchivistError):
    """Illegal REST verb (501) or option"""


class ArchivistHeaderError(ArchivistError):
    """When the expected header is not received"""


class ArchivistUnavailableError(ArchivistError):
    """Service is unavailable (503)"""


class Archivist5xxError(ArchivistError):
    """Any other 5xx error"""


def __identity(response: Response) -> str:
    identity = "unknown"
    if response.request:
        LOGGER.debug("Request %s", response.request)
        req = response.request
        body = getattr(req, "body", None)
        if body:
            # when uploading a file the body attribute is a
            # MultiPartEncoder
            try:
                body = json.loads(body)
            except (TypeError, json.decoder.JSONDecodeError):
                pass
            else:
                identity = body.get("identity", "unknown")

    return identity


def __description(response: Response) -> str:
    status_code = response.status_code
    if status_code == 404:
        return f"{__identity(response)} not found ({status_code})"

    text = response.text or ""
    url = getattr(response, "url", "")
    return f"{url}: {text} ({status_code})"


def _parse_response(response: Response):
    """Parse REST response

    Validates REST response. This is a convenience function called
    by all REST calls.

    Args:
         response (response): response from underlying REST call

    Returns:
         suitable exception if validation fails, None otherwise

    """

    status_code = response.status_code
    LOGGER.debug("Status %s", status_code)
    if status_code < 400:
        return None

    desc = __description(response)

    if status_code == 429:
        return ArchivistTooManyRequestsError(
            _headers_get(response.headers, HEADERS_RETRY_AFTER),
            desc,
        )

    if 400 <= status_code < 500:
        err, arg = {
            400: (ArchivistBadRequestError, desc),
            401: (ArchivistUnauthenticatedError, desc),
            402: (ArchivistPaymentRequiredError, desc),
            403: (ArchivistForbiddenError, desc),
            404: (ArchivistNotFoundError, desc),
        }.get(status_code, (Archivist4xxError, desc))
        return err(arg)

    if 500 <= status_code < 600:
        err, arg = {
            501: (ArchivistNotImplementedError, desc),
            503: (ArchivistUnavailableError, desc),
        }.get(status_code, (Archivist5xxError, desc))
        return err(arg)

    return ArchivistError(desc)



"""Archivist Compliance Policy Requests

   Dataclasses that represent the different types of compliance policies

"""


from dataclasses import asdict, dataclass

from .compliance_policy_type import CompliancePolicyType
from .or_dict import and_list

# NB: the order of the fields is important. Fields with default values must
#    appear after fields without. This is why the compliance_type is last
#    in every case.


@dataclass(frozen=True)
class CompliancePolicyBase:
    """
    Compliance policy base definition
    """

    description: str
    display_name: str
    asset_filter: "list[list]"

    def dict(self):
        """Emit dictionary representation"""
        d = asdict(self)
        d["asset_filter"] = and_list(d["asset_filter"])
        return d


@dataclass(frozen=True)
class CompliancePolicySince(CompliancePolicyBase):
    """
    Compliance policy that indicates if an event has 'expired'
    """

    event_display_type: str
    time_period_seconds: int
    compliance_type: str = CompliancePolicyType.COMPLIANCE_SINCE.name


@dataclass(frozen=True)
class CompliancePolicyCurrentOutstanding(CompliancePolicyBase):
    """
    Compliance policy that indicates if an event has been 'closed'
    """

    event_display_type: str
    closing_event_display_type: str
    compliance_type: str = CompliancePolicyType.COMPLIANCE_CURRENT_OUTSTANDING.name


@dataclass(frozen=True)
class CompliancePolicyPeriodOutstanding(CompliancePolicyBase):
    """
    Compliance policy that indicates if an event has been 'closed' within
    a specified time
    """

    event_display_type: str
    closing_event_display_type: str
    time_period_seconds: int
    compliance_type: str = CompliancePolicyType.COMPLIANCE_PERIOD_OUTSTANDING.name


@dataclass(frozen=True)
class CompliancePolicyDynamicTolerance(CompliancePolicyBase):
    """
    Compliance policy that indicates if the average time between opening
    and closing events in a specified period of time does not exceed a
    specified number of standard deviations from the mean.
    """

    event_display_type: str
    closing_event_display_type: str
    dynamic_window: int
    dynamic_variability: float
    compliance_type: str = CompliancePolicyType.COMPLIANCE_DYNAMIC_TOLERANCE.name


@dataclass(frozen=True)
class CompliancePolicyRichness(CompliancePolicyBase):
    """
    Compliance policy that indicates if an asset has an attribute that
    complies with a set of assertions.
    """

    richness_assertions: "list[list]"
    compliance_type: str = CompliancePolicyType.COMPLIANCE_RICHNESS.name

    def dict(self):
        """Emit dictionary representation"""
        d = asdict(self)
        d["asset_filter"] = and_list(d["asset_filter"])
        d["richness_assertions"] = and_list(d["richness_assertions"])
        return d



"""Archivist constants

"""

# these are in separate file to stop import loops
ROOT = "archivist"
SEP = "/"
VERBSEP = ":"

# These are now hardcoded and not user-selectable. Eventually they will be removed from
# the backend API and removed from this package.
ASSET_BEHAVIOURS = [
    "RecordEvidence",
]

# define in MIME canonical form
HEADERS_REQUEST_TOTAL_COUNT = "X-Request-Total-Count"
HEADERS_TOTAL_COUNT = "X-Total-Count"
HEADERS_RETRY_AFTER = "Archivist-Rate-Limit-Reset"

CONFIRMATION_STATUS = "confirmation_status"
CONFIRMATION_PENDING = "PENDING"
CONFIRMATION_FAILED = "FAILED"
CONFIRMATION_CONFIRMED = "CONFIRMED"

APPIDP_SUBPATH = "iam/v1"
APPIDP_LABEL = "appidp"
APPIDP_TOKEN = "token"

APPLICATIONS_SUBPATH = "iam/v1"
APPLICATIONS_LABEL = "applications"
APPLICATIONS_REGENERATE = "regenerate-secret"

ASSETS_SUBPATH = "v2"
ASSETS_LABEL = "assets"
ASSETS_WILDCARD = "assets/-"
EVENTS_LABEL = "events"

PUBLICASSETS_LABEL = "publicassets"

ATTACHMENTS_SUBPATH = "v1"
ATTACHMENTS_LABEL = "blobs"

ASSETATTACHMENTS_SUBPATH = "v2"
ASSETATTACHMENTS_LABEL = "attachments"

ACCESS_POLICIES_SUBPATH = "iam/v1"
ACCESS_POLICIES_LABEL = "access_policies"

COMPLIANCE_SUBPATH = "v1"
COMPLIANCE_LABEL = "compliance"

COMPLIANCE_POLICIES_SUBPATH = "v1"
COMPLIANCE_POLICIES_LABEL = "compliance_policies"

LOCATIONS_SUBPATH = "v2"
LOCATIONS_LABEL = "locations"

SBOM_PACKAGE = "Software Package"
SBOM_RELEASE = "Software Package Release"

SUBJECTS_SUBPATH = "iam/v1"
SUBJECTS_LABEL = "subjects"
SUBJECTS_SELF_ID = "subjects/00000000-0000-0000-0000-000000000000"

TENANCIES_SUBPATH = "v1"
TENANCIES_LABEL = "tenancies"

# values of tenant identity in response from other endpoints is 'tenant'
# and not 'tenancies'.
TENANCIES_PREFIX = "tenant"



"""Timestamp manipulation

   Defines timestamp policy for Archivist.

"""

from datetime import datetime

from iso8601 import parse_date
from rfc3339 import rfc3339


def parse_timestamp(date_string: str):
    """Parse an Archivist timestamp to a datetime object

    See https://pypi.org/project/iso8601/

    Args:
        date_string (str): a string representing date and time in ISO8601 format.

    Returns:
        datetime object

    """
    return parse_date(date_string)


def make_timestamp(date_object: datetime):
    """Format a datetime object into an Archivist format timestamp string

    See https://pypi.org/project/rfc3339/

    Args:
        datetime (datetime): datetime object

    Returns:
        string representation of time

    """
    return rfc3339(date_object, utc=True, use_system_timezone=False)


def now_timestamp():
    """Return Now as an Archivist format timestamp string

    See https://pypi.org/project/rfc3339/

    Returns:
        string representation of now

    """
    return make_timestamp(datetime.now())



"""Attachments interface

   Direct access to the attachments endpoint

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      with open("something.jpg") as fd:
          attachment = arch.attachments.upload(fd)

"""

# pylint:disable=too-few-public-methods


from copy import deepcopy
from io import BytesIO
from logging import getLogger
from os import path
from typing import TYPE_CHECKING, Any, BinaryIO

if TYPE_CHECKING:
    from requests.models import Response

    # pylint:disable=cyclic-import      # but pylint doesn't understand this feature
    from .archivist import Archivist

from .constants import (
    ATTACHMENTS_LABEL,
    ATTACHMENTS_SUBPATH,
)
from .dictmerge import _deepmerge
from .utils import get_url

LOGGER = getLogger(__name__)


class Attachment(dict):
    """Attachment

    Attachment object has dictionary attributes.

    """


class _AttachmentsClient:
    """AttachmentsClient

    Access to attachments entities using CRUD interface. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._subpath = f"{archivist_instance.root}/{ATTACHMENTS_SUBPATH}"
        self._label = f"{self._subpath}/{ATTACHMENTS_LABEL}"

    def __str__(self) -> str:
        return f"AttachmentsClient({self._archivist.url})"

    def get_default_key(self, data: "dict[str, str]") -> str:
        """
        Return a key to use if no key was provided
        either use filename or url as one of them is required
        """
        attachment_key = (
            data.get("filename", "")
            if data.get("filename", "")
            else data.get("url", "")
        )
        return attachment_key.replace(".", "_")

    def create(self, data: "dict[str, Any]") -> "dict[str, Any]":  # pragma: no cover
        """
        Create an attachment and return struct suitable for use in an asset
        or event creation.

        Args:
            data (dict): dictionary

        A YAML representation of the data argument would be:

            .. code-block:: yaml

                filename: functests/test_resources/doors/assets/gdn_front.jpg
                content_type: image/jpg
                display_name: arc_primary_image

            OR

            .. code-block:: yaml

                url: https://secure.eicar.org/eicar.com.zip"
                content_type: application/zip
                display_name: Test malware

             Either 'filename' or 'url' is required.
             'content_type' is required.

        Returns:

            A dict suitable for adding to an asset or event creation

        A YAML representation of the result would be:

            .. code-block:: yaml

                arc_display_name: Telephone
                arc_blob_identity: blobs/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
                arc_blob_hash_alg: SHA256
                arc_blob_hash_value: xxxxxxxxxxxxxxxxxxxxxxx
                arc_file_name: gdn_front.jpg

        """
        result = None
        file_part = None
        filename = data.get("filename")
        if filename is not None:
            _, file_part = path.split(filename)
            with open(filename, "rb") as fd:
                attachment = self.upload(fd, mtype=data.get("content_type"))

        else:
            url = data["url"]
            fd = BytesIO()
            get_url(url, fd)
            attachment = self.upload(fd, mtype=data.get("content_type"))

        result = {
            "arc_attribute_type": "arc_attachment",
            "arc_blob_identity": attachment["identity"],
            "arc_blob_hash_alg": attachment["hash"]["alg"],
            "arc_blob_hash_value": attachment["hash"]["value"],
        }

        if file_part:
            result["arc_file_name"] = file_part

        display_name = data.get("display_name")
        if display_name is not None:
            result["arc_display_name"] = display_name

        return result

    def upload(self, fd: BinaryIO, *, mtype: "str|None" = None) -> Attachment:
        """Create attachment

        Creates attachment from opened file or other data source.

        Args:
            fd (file): opened file descriptor or other file-type iterable.
            mtype (str): mimetype of data.

        Returns:
            :class:`Attachment` instance

        """

        LOGGER.debug("Upload Attachment")
        return Attachment(
            **self._archivist.post_file(
                self._label,
                fd,
                mtype,
            )
        )

    def __params(self, params: "dict[str, Any]|None") -> "dict[str, Any]":
        params = deepcopy(params) if params else {}
        # pylint: disable=protected-access
        return _deepmerge(self._archivist.fixtures.get(ATTACHMENTS_LABEL), params)

    def download(
        self,
        identity: str,
        fd: BinaryIO,
        *,
        params: "dict[str, Any]|None" = None,
    ) -> "Response":
        """Read attachment

        Reads attachment into data sink (usually a file opened for write)..
        Note that returns the response as the body will be consumed by the
        fd iterator

        Args:
            identity (str): attachment identity e.g. blobs/xxxxxxxxxxxxxxxxxxxxxxx
            fd (file): opened file descriptor or other file-type sink..
            params (dict): e.g. {"allow_insecure": "true"} OR {"strict": "true" }

        Returns:
            JSON as dict

        """
        return self._archivist.get_file(
            f"{self._subpath}/{identity}",
            fd,
            params=self.__params(params),
        )

    def info(
        self,
        identity: str,
    ) -> "dict[str, Any]":
        """Read attachment info

        Reads attachment info

        Args:
            identity (str): attachment identity e.g. blobs/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            REST response

        """
        return self._archivist.get(f"{self._subpath}/{identity}/info")



"""assets confirmer interface
"""


from logging import getLogger
from typing import TYPE_CHECKING

import backoff

if TYPE_CHECKING:
    from .subjects import Subject, _SubjectsClient

from .constants import (
    CONFIRMATION_CONFIRMED,
    CONFIRMATION_STATUS,
)
from .errors import ArchivistUnconfirmedError

# pylint:disable=cyclic-import      # but pylint doesn't understand this feature
from .utils import backoff_handler

MAX_TIME = 1200

LOGGER = getLogger(__name__)


def __lookup_max_time():
    return MAX_TIME


def __on_giveup_confirmation(details):
    identity = details["args"][1]
    elapsed = details["elapsed"]
    raise ArchivistUnconfirmedError(
        f"confirmation for {identity} timed out after {elapsed} seconds"
    )


@backoff.on_predicate(
    backoff.expo,
    logger=None,  # pyright: ignore
    max_time=__lookup_max_time,
    on_backoff=backoff_handler,
    on_giveup=__on_giveup_confirmation,
)
def _wait_for_confirmation(self: "_SubjectsClient", identity: str) -> "Subject":
    """Return None until subjects is confirmed"""
    subject = self.read(identity)
    if CONFIRMATION_STATUS not in subject:
        return None  # pyright: ignore

    if subject[CONFIRMATION_STATUS] == CONFIRMATION_CONFIRMED:
        return subject

    return None  # pyright: ignore



"""Archivist OR dictionary and list

   Used in filters: list of Or dictionaries

   Dictionaries where key is always 'or' and value is a list of strings
"""


from typing import Any


def or_dict(list_: list) -> "dict[str, list]":
    """Construct a dictionary with key 'or'"""
    return {"or": list_}


def and_list(lists: list) -> "list[dict[str, list[Any]]]":
    """Construct a list of or dictionaries"""
    return [or_dict(j) for j in lists]



"""Asset data class

"""
from contextlib import suppress


class Asset(dict):
    """Asset

    Asset object has dictionary attributes and properties.

    """

    @property
    def primary_image(self) -> "str|None":
        """Primary Image

        Attachment that is the primary image of the asset.

        Returns:
            :class:`Attachment` instance

        """

        # try latest API
        try:
            return self["attributes"]["arc_primary_image"]
        except (KeyError, TypeError):
            pass

        # try old API
        try:
            attachments = self["attributes"]["arc_attachments"]
        except (KeyError, TypeError):
            pass
        else:
            return next(  # pragma: no cover
                (
                    a
                    for a in attachments
                    if "arc_display_name" in a
                    if a["arc_display_name"] == "arc_primary_image"
                ),
                None,
            )

        return None

    @property
    def name(self) -> "str|None":
        """str: name of the asset"""
        name = None
        with suppress(KeyError, TypeError):
            name = self["attributes"]["arc_display_name"]

        return name



"""common parser argument
"""

# pylint:  disable=missing-docstring
# pylint:  disable=too-few-public-methods


import argparse
from enum import Enum
from logging import getLogger
from sys import exit as sys_exit
from warnings import filterwarnings

from .archivist import Archivist
from .dictmerge import _deepmerge
from .logger import set_logger
from .proof_mechanism import ProofMechanism
from .utils import get_auth

filterwarnings("ignore", message="Unverified HTTPS request")


LOGGER = getLogger(__name__)


# from https://stackoverflow.com/questions/43968006/support-for-enum-arguments-in-argparse
class EnumAction(argparse.Action):
    """
    Argparse action for handling Enums
    """

    def __init__(self, **kwargs):
        # Pop off the type value
        enum_type = kwargs.pop("type", None)

        # Ensure an Enum subclass is provided
        if enum_type is None:
            raise ValueError("type must be assigned an Enum when using EnumAction")

        if not issubclass(enum_type, Enum):
            raise TypeError("type must be an Enum when using EnumAction")

        # Generate choices from the Enum
        kwargs.setdefault("choices", tuple(e.name for e in enum_type))

        super().__init__(**kwargs)

        self._enum = enum_type

    def __call__(self, parser, namespace, values, option_string=None):
        # Convert value back into an Enum
        value = self._enum[values]  # pyright: ignore
        setattr(namespace, self.dest, value)


def simple_parser(description: str):
    """Construct parser with bare minimum features"""
    parser = argparse.ArgumentParser(
        description=description,
    )
    return parser


def common_parser(description: str):
    """Construct parser with security option for token/auth authentication"""
    parser = simple_parser(description)
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="print verbose debugging",
    )
    parser.add_argument(
        "-u",
        "--url",
        type=str,
        dest="url",
        action="store",
        default="https://app.datatrails.ai",
        help="url of Archivist service",
    )
    parser.add_argument(
        "-p",
        "--proof-mechanism",
        type=ProofMechanism,
        action=EnumAction,
        dest="proof_mechanism",
        default=None,
        help="mechanism for proving the evidence for events on the Asset",
    )
    parser.add_argument(
        "--auth-token",
        type=str,
        dest="auth_token",
        action="store",
        default=None,
        help="API token value",
    )
    parser.add_argument(
        "--auth-token-filename",
        type=str,
        dest="auth_token_filename",
        action="store",
        default=None,
        help="FILE containing API authentication token",
    )
    parser.add_argument(
        "--client-id",
        type=str,
        dest="client_id",
        action="store",
        default=None,
        help="Client ID from appregistrations",
    )
    parser.add_argument(
        "--client-secret",
        type=str,
        dest="client_secret",
        action="store",
        default=None,
        help="Client secret from appregistrations",
    )
    parser.add_argument(
        "--client-secret-filename",
        type=str,
        dest="client_secret_filename",
        action="store",
        default=None,
        help="FILE containing client secret from appregistrations",
    )
    parser.add_argument(
        "-n",
        "--namespace",
        type=str,
        dest="namespace",
        action="store",
        default=None,
        help="namespace of item population",
    )

    return parser


def endpoint(args):
    if args.verbose:
        set_logger("DEBUG")
    else:
        set_logger("INFO")

    arch = None
    LOGGER.info("Initializing connection to DATATRAILS...")
    fixtures = {}
    if args.proof_mechanism is not None:
        fixtures = {
            "assets": {
                "proof_mechanism": args.proof_mechanism.name,
            },
        }

    if args.namespace is not None:
        fixtures = _deepmerge(
            fixtures,
            {
                "assets": {
                    "attributes": {
                        "arc_namespace": args.namespace,
                    },
                },
                "locations": {
                    "attributes": {
                        "namespace": args.namespace,
                    },
                },
            },
        )

    auth = get_auth(
        auth_token=args.auth_token,
        auth_token_filename=args.auth_token_filename,
        client_id=args.client_id,
        client_secret=args.client_secret,
        client_secret_filename=args.client_secret_filename,
    )

    if auth is None:
        LOGGER.error("Critical error.  Aborting.")
        sys_exit(1)

    arch = Archivist(args.url, auth, fixtures=fixtures)
    if arch is None:
        LOGGER.error("Critical error.  Aborting.")
        sys_exit(1)

    return arch
